#!/usr/bin/python

import bzrlib.tests
import os.path

class ShelfTests(bzrlib.tests.TestCaseWithTransport):
    ORIGINAL = '\n\nhello test world\n\n'
    MODIFIED = '\n\ngoodbye test world\n\n'
    DIFF_HEADER = "=== modified file 'test_file'\n"
    DIFF_1 = """--- test_file\t
+++ test_file\t
@@ -1,4 +1,4 @@
 
 
-hello test world
+goodbye test world
 
"""
    DIFF_2 = """--- test_file\t
+++ test_file\t
@@ -1,4 +1,4 @@
 
 
-goodbye test world
+hello test world
 
"""
    def test_shelf(self):
        self.__test_loop(1)

    def test_shelf_multi(self):
        self.__test_loop(10)

    def __test_loop(self, count):
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)

        while count > 0:
            count -= 1

            # Modify the test file
            file('test_file', 'w').write(self.MODIFIED)

            # Check the diff is right
            self.assertEqual(self.capture('diff', retcode=1),
                self.DIFF_HEADER + self.DIFF_1 + '\n')

            # Shelve the changes
            self.run_bzr('shelve', retcode=0)

            # Make sure there is no diff anymore
            self.assertEqual(self.capture('diff', retcode=0), '')

            # Make sure the file is actually back the way it was
            self.assertEqual(file('test_file').read(), self.ORIGINAL)

            # Check the shelf is right
            shelf = open(os.path.join(tree.branch.base,
                        '.shelf/shelves/default/00')).read()
            shelf = shelf[shelf.index('\n') + 1:] # skip the message
            self.assertEqual(shelf, self.DIFF_1)

            # Unshelve
            self.run_bzr('unshelve', retcode=0)

            # Check the diff is right again
            self.assertEqual(self.capture('diff', retcode=1),
                self.DIFF_HEADER + self.DIFF_1 + '\n')

            # Make sure the file is back the way it should be
            self.assertEqual(file('test_file').read(), self.MODIFIED)

    def test_shelf_nothing_to_shelve(self):
        import os.path
        tree = self.make_branch_and_tree('.')
        self.__create_and_add_test_file(tree)

        # Shelve the changes
        self.run_bzr('shelve', retcode=3)

        if os.path.exists(os.path.join(tree.branch.base,
                '.shelf/shelves/default/00')):
            self.fail("Shelf exists, but it shouldn't")

    def __create_and_add_test_file(self, tree):
        self.build_tree_contents([('test_file', self.ORIGINAL)])
        tree.add('test_file')
        tree.commit(message='add test_file')

    def test_shelf_with_revision(self):
        tree = self.make_branch_and_tree('.')

        self.__create_and_add_test_file(tree)

        # Modify the test file and commit it
        self.build_tree_contents([('test_file', self.MODIFIED)])
        tree.commit(message='update test_file')

        # Shelve the changes
        self.run_bzr('shelve', '-r', '1', retcode=0)

        # Check the diff is right
        self.assertEqual(self.capture('diff', retcode=1),
            self.DIFF_HEADER + self.DIFF_2 + '\n')

        # Make sure the file is the way it should be
        self.assertEqual(file('test_file').read(), self.ORIGINAL)

        # Unshelve
        self.run_bzr('unshelve', retcode=0)

        # Make sure the file is back the way it should be
        self.assertEqual(file('test_file').read(), self.MODIFIED)

    def test_shelf_with_two_revisions(self):
        tree = self.make_branch_and_tree('.')

        cmd = 'shelve -r 1..2'
        (stdout, stderr) = self.run_bzr_captured(cmd.split(), retcode=None)

        self.assertEqual(stderr.split('\n')[0],
            'bzr: ERROR: shelve only accepts a single revision parameter.')

    def test_shelf_long_filename(self):
        """Regression test for diffstat with long filenames.

        Create a branch with two files, one of which has a long name. Commit.
        Modify both files. Shelve the file with the short name. If diffstat
        has regressed, it will generate a diffstat of the file with the long
        name, and break.
        """
        self.run_bzr('init')
        filename = 'a' * 80
        file(filename, 'wb').write('hello')
        file('foo', 'wb').write('bar')
        self.run_bzr('add')
        self.run_bzr('commit', '-m', 'added')
        file(filename, 'wb').write('goodbye')
        file('foo', 'wb').write('baz')
        self.run_bzr('shelve', 'foo', retcode=0)

